/*************************************************************************
 * Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.
 *
 * See LICENSE.txt for license information
 ************************************************************************/

#ifndef PLUGIN_H_
#define PLUGIN_H_

int exampleProfilerStart(int eActivationMask);
int exampleProfilerStop(void);

#endif
